/****************************************************************************
 Include Section
*****************************************************************************/

#include "global.h"

/*****************************************************************************
 Symbolic constants / definitions
******************************************************************************/

#define TWI0SCL                 PE14
#define TWI0SDA                 PE15
#define TWI_DCNT_VAL            1
#define BF_I2C_BOOT_ADDRESS     0x5F

/*****************************************************************************
 Global Variables
******************************************************************************/

.EXTERN loader_file;

/*****************************************************************************
 Variables
******************************************************************************/

.SECTION L1_data;
.ALIGN 4;

/*****************************************************************************
 Functions
******************************************************************************/

.SECTION L1_code;

/*****************************************************************************
 Prototypes
******************************************************************************/

.GLOBAL __twi_hostboot_run;

//.GLOBAL gpio_ports_setup;
//.GLOBAL twi_reset;
//.GLOBAL twi_setup;
//.GLOBAL twi_putc;
//.GLOBAL twi_disable;


/****************************************************************************
 Function:              __twi_hostboot_run
 Description:           do the transfer
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7,P5:3
 Global Registers Used: none
 C-Callable :           no
*****************************************************************************/

__twi_hostboot_run:

    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:7,P5:3);
    PUSH_LOOP_CNTx(0);
    IMM32(P5,SYS_MMR_BASE);


    call gpio_ports_setup;

    /* wait for slave to deassert host wait signal */
    twi_hostboot_run_NI.initial_wait:
    R7 = w[P5 + lo(PORTE)] (z);
    CC = bittst(R7,bitpos(PFHW));
    if CC jump twi_hostboot_run_NI.initial_wait;

    call twi_reset;
    call twi_setup;

    /* This is our pointer to the LDR file */
    IMM32(P4,loader_file);
    IMM32(P3,length(loader_file));

    /***********************************************************
     Pre-load the tx FIFO with the first two bytes: this is
     necessary to avoid the generation of the Buffer Read Error
     (BUFRDERR) which occurs whenever a transmit transfer is
     initiated while the transmit buffer is empty
    ***********************************************************/
    R7 = b[P4++](z);
    w[P5 + lo(TWI0_XMT_DATA8)] = R7;


    /***********************************************************
     Initiating the Write operation
     Program the HOST Control register with:
     1. the number of bytes to transfer: TWICount(x)
     2. Repeated Start (RESTART): optional
     3. speed mode: FAST or Standard
     4. direction of transfer:
     MDIR = 1 for reads, MDIR = 0 for writes
     5. HOST Enable MEN. Setting this bit will kick off the transfer
    ***********************************************************/
    #if (FSCL_kHz > 100)
    R7 =   nSCLOVR                      |   /* Serial Clock Override */
           nSDAOVR                      |   /* Serial Data Override */
            SET_TWI_DCNT(TWI_DCNT_VAL)  |   /* Data Transfer Count */
            RSTART                      |   /* Repeat Start */
           nSTOP                        |   /* Issue Stop Condition */
            FAST                        |   /* Fast Mode */
            MEN                         (z);/* HOST Mode Enable */
    #else
    R7 =   nSCLOVR                      |   /* Serial Clock Override */
           nSDAOVR                      |   /* Serial Data Override */
            SET_TWI_DCNT(TWI_DCNT_VAL)  |   /* Data Transfer Count */
            RSTART                      |   /* Repeat Start */
           nSTOP                        |   /* Issue Stop Condition */
           nFAST                        |   /* Fast Mode */
           nMDIR                        |   /* HOST Transfer Direction */
            MEN                         (z);/* HOST Mode Enable */
    #endif
    w[P5 + lo(TWI0_MASTER_CTL)] = R7;
    ssync;
    P3 += -1;   /* one byte already sent out */


    lsetup (twi_hostboot_run.loop_start,twi_hostboot_run.loop_end) LC0 = P3;
    twi_hostboot_run.loop_start:    R7 = w[P5 + lo(TWI0_INT_STAT)](z);
                                    CC = bittst(R7, bitpos(XMTSERV));
                                    if !CC jump twi_hostboot_run.loop_start;
                                    w[P5 + lo(TWI0_INT_STAT)] = R7;

                                    /* wait for slave to deassert host wait signal */
                                    twi_hostboot_run.loop_wait:
                                    R7 = w[P5 + lo(PORTE)] (z);
                                    CC = bittst(R7,bitpos(PFHW));
                                    if CC jump twi_hostboot_run.loop_wait;

                                    /* reload transmit buffer */
                                    R7 = b[P4++](z);
                                    w[P5 + lo(TWI0_XMT_DATA8)] = R7;
                                    ssync;

                                    /* kick off transfer */
                                    R7 = SET_TWI_DCNT(TWI_DCNT_VAL);
                                    R6 = w[P5 + lo(TWI0_MASTER_CTL)];
                                    R7 = R7 | R6;
                                    w[P5 + lo(TWI0_MASTER_CTL)] = R7;
    twi_hostboot_run.loop_end:      ssync;


    /* check that HOST transfer has completed */
    M_COMP: R7 = w[P5 + lo(TWI0_INT_STAT)](z);
            CC = bittst(R7, bitpos(MCOMP));
            if !CC jump M_COMP;
            w[P5 + lo(TWI0_INT_STAT)] = R7;


    call twi_disable;


    POP_LOOP_CNTx(0);
    (R7:7,P5:3) = [SP++];
    ASTAT = [SP++];
    unlink;

    rts;
__twi_hostboot_run.END:


/*****************************************************************************
 Function:              twi_putc
 Description:           run a SPI tranmission -> transmit SPI
                        Hardware Loop used
                        No Interrupts
 Input Parameters:      R0
 Return Parameters:     none
 Registers Used:        R7:6,ASTAT
 Global Registers Used: P5 -> assumed to contain SYS_MMR_BASE address
 C-Callable :           no
*****************************************************************************/

twi_putc:

    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:6);

    w[P5 + lo(TWI0_XMT_DATA8)] = R0;
    ssync;
    R7 = SET_TWI_DCNT(TWI_DCNT_VAL);
    R6 = w[P5 + lo(TWI0_MASTER_CTL)];
    R7 = R7 | R6;
    w[P5 + lo(TWI0_MASTER_CTL)] = R7;
    ssync;

    (R7:6) = [SP++];
    ASTAT = [SP++];
    unlink;

    rts;
twi_putc.end:


/****************************************************************************
 Function:              gpio_ports_setup
 Description:           setup up GP ports
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7:6,P5
 Global Registers Used: none
 C-Callable :           no
*****************************************************************************/

gpio_ports_setup:

    link 0;
    [--SP] = (R7:6,P5:5);
    IMM32(P5,SYS_MMR_BASE);


    R7 = w[P5 + lo(PORTE_FER)];
    R6 = ( TWI0SCL | TWI0SDA )(z);
    R7 = R7 | R6;
    w[P5 + lo(PORTE_FER)] = R7;

    /* enable TWI0_SCL and TWI0_SDA mux mode */
    R7 = [P5 + lo(PORTE_MUX)];
    IMM32(R6,MUX(0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0));
    R7 = R7 | R6;
    [P5 + lo(PORTE_MUX)] = R7;


    /* Set GPIO direction to input */
    R7 = PFHW;
    w[P5 + lo(PORTE_DIR_CLEAR)] = R7;

    /* Enable PF */
    R7 = w[P5 + lo(PORTE_INEN)](z);
    bitset(R7,bitpos(PFHW));
    w[P5 + lo(PORTE_INEN)] = R7;


    (R7:6,P5:5) = [SP++];
    unlink;

    rts;
gpio_ports_setup.end:


/****************************************************************************
 Function:              twi_reset
 Description:           reset TWI controller
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7,P5
 Global Registers Used: none
 C-Callable :           no
*****************************************************************************/

twi_reset:

    link 0;
    [--SP] = (R7:7,P5:5);
    IMM32(P5,SYS_MMR_BASE);

    /* Reset the TWI controller */
    R7 = 0;
    w[P5 + lo(TWI0_CONTROL)] = R7;
    
    /* Clearing all status bits */
    R7 =   nBUSBUSY     |   /* Bus Busy */
           nSCLSEN      |   /* Serial Clock Sense */
           nSDASEN      |   /* Serial Data Sense */
            BUFWRERR    |   /* Buffer Write Error */
            BUFRDERR    |   /* Buffer Read Error */
            DNAK        |   /* Data Not Acknowledged */
            ANAK        |   /* Address Not Acknowledged */
            LOSTARB     |   /* Lost Arbitration Indicator (Xfer Aborted) */
           nMPROG       (z);/* HOST Transfer in Progress */
    w[P5 + lo(TWI0_MASTER_STAT)] = R7;
   
    /* Clearing all pending interrupt flags */
    R7 =    SCLI        |   /* Serial Clock Interupt */
            SDAI        |   /* Serial Data Interrupt */
            RCVSERV     |   /* Receive FIFO Service */
            XMTSERV     |   /* Transmit FIFO Service */
            MERR        |   /* HOST Transfer Error */
            MCOMP       |   /* HOST Transfer Complete */
            SOVF        |   /* Slave Overflow */
            SERR        |   /* Slave Transfer Error */
            SCOMP       |   /* Slave Transfer Complete */
            SINIT       (z);/* Slave Transfer Initiated */
   w[P5 + lo(TWI0_INT_STAT)] = R7;

    /* flush the buffers */
    R7 =   nRCVINTLEN   |   /* Receive Buffer Interrupt Length */
           nXMTINTLEN   |   /* Transmit Buffer Interrupt Length */
            RCVFLUSH    |   /* Receive Buffer Flush */
            XMTFLUSH    (z);/* Transmit Buffer Flush */
    w[P5 + lo(TWI0_FIFO_CTL)] = R7;


    ssync;
    (R7:7,P5:5) = [SP++];
    unlink;

    rts;
twi_reset.END:


/****************************************************************************
 Function:              twi_setup
 Description:           generic TWI setup up function
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7:R0,P5
 Global Registers Used: none
 C-Callable :           no
*****************************************************************************/

twi_setup:

    link 0;
    [--SP] = (R7:0,P5:5);
    IMM32(P5,SYS_MMR_BASE);


    /* Write to TWI_CONTROL to set prescale and enable the TWI */
    call __get_sclk_hz;
    IMM32(R1,10000000); /* 10MHz time reference, period = 100ns */
    udiv32(); /* R0 = PRESCALE = SCLK/10MHz */
    R1 = TWI_ENA(z);
    R0 = R0 | R1;
    w[P5 + lo(TWI0_CONTROL)] = R0;


    /* Write to TWI_CLK_DIV */
    /* calculate TWI CLKHI/CLKLOW */
    /* CLKHI : CLKLO ratio is fixed to 1:1 */
    R0 = 10000;
    R1 = 2*FSCL_kHz;
    udiv32(); /* R0 = CLKLO */

    R7 = 8;
    R1 = R0;
    R1 <<= R7; /* R1 = CLKHI */
    R0 = R0 | R1;
    w[P5 + lo(TWI0_CLKDIV)] = R0;


    /* Wite to TWI_HOST_ADDR with the Address of the targeted device */
    R7 = SET_TWI_ADDR(BF_I2C_BOOT_ADDRESS);
    w[P5 + lo(TWI0_MASTER_ADDR)] = R7;


    /* Write to TWI_FIFO_CTL to select whether 1 or 2 Bytes generate interrupts */
    /* clear the flush bits */
    R7 = 0;
    w[P5 + lo(TWI0_FIFO_CTL)] = R7;


    /* Write to TWI_INT_MASK to unmask TWI events to generate interrupts */
    R7 =   nSCLIM       |   /* Serial Clock Interupt */
           nSDAIM       |   /* Serial Data Interrupt */
           nRCVSERVM    |   /* Receive FIFO Service Interrupt Mask */
            XMTSERVM    |   /* Transmit FIFO Service Interrupt Mask */
           nMERRM       |   /* HOST Transfer Error Interrupt Mask */
           nMCOMPM      |   /* HOST Transfer Complete Interrupt Mask */
           nSOVFM       |   /* Slave Overflow Interrupt Mask */
           nSERRM       |   /* Slave Transfer Error Interrupt Mask */
           nSCOMPM      |   /* Slave Transfer Complete Interrupt Mask */
           nSINITM      (z);/* Slave Transfer Initiated Interrupt Mask */
    w[P5 + lo(TWI0_INT_MASK)] = R7;


    ssync;
    (R7:0,P5:5) = [SP++];
    unlink;

    rts;
twi_setup.end:


/****************************************************************************
 Function:              twi_disable
 Description:           generic TWI disable function
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7,P5,ASTAT
 Global Registers Used: none
 C-Callable :           no
*****************************************************************************/

twi_disable:

    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:7,P5:5);
    IMM32(P5,SYS_MMR_BASE);


    twi_disable.wait:
    R7 = w[P5 + lo(TWI0_FIFO_STAT)];
    CC = R7 == 0;
    if !CC jump twi_disable.wait;

    call twi_reset;
    R7 = 0;
    w[P5 + lo(TWI0_CONTROL)] = R7;
    w[P5 + lo(TWI0_CLKDIV)] = R7;
    w[P5 + lo(TWI0_MASTER_CTL)] = R7;
    w[P5 + lo(TWI0_MASTER_ADDR)] = R7;
    w[P5 + lo(TWI0_FIFO_CTL)] = R7;
    w[P5 + lo(TWI0_INT_MASK)] = R7;


    ssync;
    (R7:7,P5:5) = [SP++];
    ASTAT = [SP++];
    unlink;

    rts;
twi_disable.end:


/*****************************************************************************
 EOF
******************************************************************************/
